"""Transformer module for the transformation of lark trees."""

import typing
import logging

from lark.tree import Tree
from lark import Transformer
from lark.lexer import Token
from collections import defaultdict, namedtuple, OrderedDict

from ebel.cache import _BelScript
from ebel.constants import FILE, URL, PATTERN, LIST

log = logging.getLogger(__name__)


def get_token_dict2(token_tree_list: list) -> dict:
    """Create a dict out of a mixed list of lark.lexer.Token and tuple('name_of_rule',list(lark.lexer.Token,...)).

    Parameters
    ----------
    token_tree_list : list
        Description of parameter `token_tree_list`.

    Returns
    -------
    type
        Description of returned object.

    """
    ret_dict = defaultdict(list)

    for item in token_tree_list:

        if isinstance(item, Token):
            ret_dict[item.type].append(item.value)

        elif isinstance(item, tuple):
            ret_dict2 = defaultdict(list)

            for item2 in item[1]:
                if isinstance(item2, Token):
                    ret_dict2[item2.type].append(item2.value)

            ret_dict[item[0]] = dict(ret_dict2)
        elif isinstance(item, Tree):
            for item3 in item.children:
                ret_dict[item3.type].append(item3.value)

    return dict(ret_dict)


class _BelTransformer(Transformer):
    """Transformer optimized for JSON output.

    Vertices
    --------
        are tuples with 2 elements
        [0] := vertex name
        [1] := values

    Properties of vertices
    ----------------------
        are in OrderedDict's
        - properties with 1 value => namedtuple
        - properties with * values => List[namedtuple]
    """

    exclude_token_types = ('OB', 'CB', 'QM', 'COLON', 'COMMA', 'OCB', 'CCB')

    nt_frag = namedtuple('fragment', ('range', 'descriptor'))
    nt_nn = namedtuple('nn', ('namespace', 'name'))
    nt_var = namedtuple('variant', ('hgvs',))
    nt_ma = namedtuple('ma', ('namespace', 'name', 'default'))
    nt_pmod = namedtuple('protein_modification', ('namespace', 'name', 'type', 'amino_acid', 'position'))
    nt_support = namedtuple('support', ('text', ))
    nt_citation = namedtuple('citation', ('type', 'title', 'ref', 'pub_date', 'author_list', 'comment'))
    nt_type = namedtuple('function', ('type', 'name'))
    nt_gmod = namedtuple('gene_modification', ('namespace', 'name'))
    named_tuples = (nt_frag,
                    nt_nn,
                    nt_var,
                    nt_pmod,
                    nt_support,
                    nt_citation,
                    nt_type,
                    nt_ma,
                    nt_gmod)

    def __init__(self, force_new_db=False):
        Transformer.__init__(self)
        self.cache = _BelScript(force_new_db=force_new_db)
        self.citations = []

    def script(self, n):
        return n

    def statements_and_sets(self, n):
        return {'statements_and_sets': n}

    def abundance(self, n):
        """Return abundance."""
        return [self.nt_type('abundance', 'abundance'), n]

    def gene(self, n):
        """Return gene."""
        return [self.nt_type('abundance', 'gene'), n]

    def micro_rna(self, n):
        """Return micro_rna."""
        return [self.nt_type('abundance', 'micro_rna'), n]

    def rna(self, n):
        """Return rna."""
        return [self.nt_type('abundance', 'rna'), n]

    def protein(self, n):
        """Return protein."""
        return [self.nt_type('abundance', 'protein'), n]

    def population(self, n):
        """Return protein."""
        return [self.nt_type('abundance', 'population'), n]

    def composite(self, n):
        """Return composite."""
        return [self.nt_type('list', 'composite'), sorted(n)]

    def definitions(self, n):
        """Return definitions of namespace and annotations."""
        return {'definitions': n}

    def sec(self, n):
        """Return cellSecretion."""
        return [self.nt_type('transformation', 'cell_secretion'), [n[0]]]

    def _ns_anno_props(self, n):
        """Return properties of namespace or annotation."""
        token_names = ('KEYWORD', ('URL', 'PATTERN', 'LIST'), ('URL_DEF', 'PATTERN_DEF', 'FILE_PATH'))

        keyword, type_, value = self._get_values(n, token_names)

        value_list = []
        if type_ == 'LIST':
            in_list = self._get_tree(n, 'in_list')
            value_list = self._get_all_values_by_name(in_list, 'ENTRY')

        props = OrderedDict([
            ('keyword', keyword),
            ('type', type_),
            ('value', value),
            ('value_list', sorted(value_list))
        ])
        return props

    def namespace(self, n):
        """Return namespace vertex."""
        token_dict = get_token_dict2(n)

        keyword = token_dict['KEYWORD'][0]
        namespace_type, value = None, None

        if 'LIST' in token_dict:
            value = token_dict['ENTRY']
            namespace_type = LIST

        elif 'URL_DEF' in token_dict:
            value = token_dict['URL_DEF'][0]
            namespace_type = URL

        elif 'FILE_PATH' in token_dict:
            value = token_dict['FILE_PATH'][0]
            namespace_type = FILE

        elif 'PATTERN_DEF' in token_dict:
            value = token_dict['PATTERN_DEF'][0]
            namespace_type = PATTERN

        self.cache.set_namespace_definition(namespace_type, keyword, value)

        props = self._ns_anno_props(n)
        return {'namespace': props}

    def annotation(self, n):
        """Return namespace vertex."""
        token_dict = get_token_dict2(n)

        keyword = token_dict['KEYWORD'][0]
        annotation_type, value = None, None

        if 'LIST' in token_dict:
            value = token_dict['ENTRY']
            annotation_type = LIST

        elif 'URL_DEF' in token_dict:
            value = token_dict['URL_DEF'][0]
            annotation_type = URL

        elif 'FILE_PATH' in token_dict:
            value = token_dict['FILE_PATH'][0]
            annotation_type = FILE

        elif 'PATTERN_DEF' in token_dict:
            value = token_dict['PATTERN_DEF'][0]
            annotation_type = PATTERN

        self.cache.set_annotation_definition(annotation_type, keyword, value)

        props = self._ns_anno_props(n)
        return {'annotation': props}

    def document(self, n):
        """Return Document vertex."""
        return {'document': OrderedDict(sorted(n))}

    def _doc_prop(self, name, n):
        return (name, self._get_values(n, (('STRING_IN_QUOTES', 'WORD'),))[0])

    def document_name(self, n):
        """Return document property as tuple."""
        return self._doc_prop('name', n)

    def document_description(self, n):
        """Return document property as tuple."""
        return self._doc_prop('description', n)

    def document_version(self, n):
        """Return document property as tuple."""
        return self._doc_prop('version', n)

    def document_authors(self, n):
        """Return document property as tuple."""
        return self._doc_prop('authors', n)

    def document_contact_info(self, n):
        """Return document property as tuple."""
        return self._doc_prop('contact_info', n)

    def document_copyright(self, n):
        """Return document property as tuple."""
        return self._doc_prop('copyright', n)

    def document_licences(self, n):
        """Return document property as tuple."""
        return self._doc_prop('licences', n)

    def document_keywords(self, n):
        """Return document property as tuple."""
        return self._doc_prop('keywords', n)

    def deg(self, n):
        """Return degradation."""
        return [self.nt_type('transformation', 'degradation'), [n[0]]]

    def complex_abundance(self, n):
        """Return complex as abundance."""
        return [self.nt_type('abundance', 'complex'), [n[0]]]

    def complex_list(self, n):
        """Return complex as list."""
        return [self.nt_type('list', 'complex'), n]

    def list_complex(self, n):
        """Return abundance list of complex."""
        return sorted(n)

    def list(self, n):
        """Return list."""
        function_type = self._format_sub_obj([self.nt_type('list', 'list')])[0]
        return {'object': [function_type, [self._format_sub_obj(x) for x in sorted(n)]]}

    def act2(self, n):
        """Return activity."""
        abundance = self._get_tree(n, 'act_abundance').children[0]
        ma = self._get_dict_value(n, 'ma')
        if ma:
            return [self.nt_type('process', 'activity'), abundance, ma._asdict()]
        else:
            return [self.nt_type('process', 'activity'), abundance]

    def act(self, n):
        """Return activity."""
        return [self.nt_type('process', 'activity'), n]

    def tloc(self, n):
        """Return tloc."""
        return [self.nt_type('transformation', 'translocation'), n]

    def _get_rel(self, n):
        return {'relation': n[0].data}

    def nested_relation(self, n):
        """Return nested relation."""
        return {'nested_relation': n[0]}

    def relation_with_list(self, n):
        """Return relation."""
        return self._get_rel(n)

    def relation_basic(self, n):
        """Return relation."""
        return self._get_rel(n)

    def rel_with_list_object(self, n):
        """Return relation."""
        return self._get_rel(n)

    def biomarker_relation(self, n):
        """Return relation."""
        return self._get_rel(n)

    def process_relation(self, n):
        """Return relation."""
        return self._get_rel(n)

    def analogous_relation(self, n):
        """Return relation."""
        return self._get_rel(n)

    def orthologous_relation(self, n):
        """Return orthologous relation."""
        return self._get_rel(n)

    def additional_relations(self, n):
        """Return relation."""
        return self._get_rel(n)

    def transc_relation(self, n):
        """Return relation."""
        return self._get_rel(n)

    def transl_relation(self, n):
        """Return relation."""
        return self._get_rel(n)

    def has_members(self, n):
        """Return relation."""
        return {'relation': 'has_members'}

    def statement(self, n: list):
        """Return statement."""
        return {'statement': n}

    def subject(self, n):
        """Return subject."""
        return {'subject': self._format_sub_obj(n[0])}

    def nested_subject(self, n):
        """Return nested subject."""
        return {'nested_subject': self._format_sub_obj(n[0])}

    def nested_object(self, n):
        """Return nested subject."""
        return {'nested_object': self._format_sub_obj(n[0])}

    def sets(self, n):
        """Return sets."""
        return {'sets': n}

    def evidence(self, n):
        """Return support.

        support formily known as evidence, supprtingText
        """
        return {'evidence': n[0].value}

    def citation(self, n):
        """Return citation."""
        tv_dict = {x.type: x.value for x in n if isinstance(x, Token)}
        c_title, c_pubdate, c_author_list, c_comment = '', '', '', ''
        c_type = tv_dict['C_TYPE']

        if "ONLY_2_PARAMETERS" in tv_dict:
            c_ref = tv_dict['C_PARAM2']
        else:
            c_title = tv_dict['C_PARAM2']
            c_ref = tv_dict['C_PARAM3']
            c_pubdate = tv_dict['C_PUBDATE'] if 'C_PUBDATE' in tv_dict else ''
            c_author_list = tv_dict['C_AUTHORLIST'] if 'C_AUTHORLIST' in tv_dict else ''
            c_comment = tv_dict['C_COMMENT'] if 'C_COMMENT' in tv_dict else ''

        nt = self.nt_citation(type=c_type,
                              title=c_title,
                              ref=c_ref,
                              pub_date=c_pubdate,
                              author_list=c_author_list,
                              comment=c_comment)

        self.citations.append({'citation_type': c_type, 'citation_id': c_ref})

        return {'citation': nt._asdict()}

    def statement_group(self, n):
        """Return statement group."""
        return {'statement_group': self._get_value(n, 'GROUP_NAME')}

    def set_annotation(self, n):
        """Return annotation sets."""
        entries = []
        annotation_key = ""
        for tkn in n:
            if tkn.type == "KEYWORD":
                annotation_key = tkn.value
            elif tkn.type == "ANNO_SET_ENTRY":
                entries.append(tkn)

        for entry in entries:
            self.cache.set_annotation_entry(
                annotation=annotation_key,
                entry=entry.value,
                token=entry
            )

        keyword = self._get_value(n, 'KEYWORD')
        entries = self._get_all_values_by_name(n, 'ANNO_SET_ENTRY')
        return {'set': {keyword: entries}}

    def unset(self, n):
        """Return unsets."""
        keywords = self._get_all_values_by_name(n, 'ANNO_KEYWORD')
        return {'unset': keywords}

    def object(self, n):
        """Return object."""
        return {'object': self._format_sub_obj(n[0])}

    def molec_process(self, n):
        """Return molecular process."""
        return {'object': self._format_sub_obj(n[0])}

    def act_or_abundance(self, n):
        """Return activity or abundance."""
        return {'act_or_abundance': self._format_sub_obj(n[0])}

    def transformation(self, n):
        """Return transformation."""
        return {'transformation': self._format_sub_obj(n[0])}

    def pat(self, n):
        """Return object."""
        return {'subject': self._format_sub_obj(n[0])}

    def transc_subject(self, n):
        return self.subject(n)

    def transc_object(self, n):
        return self.object(n)

    def transl_subject(self, n):
        return self.subject(n)

    def transl_object(self, n):
        return self.object(n)

    def ortho_subject(self, n):
        """Return orthologous gene subject."""
        return self.subject(n)

    def ortho_object(self, n):
        """Return orthologous gene object."""
        return self.object(n)

    def surf(self, n):
        return [self.nt_type('transformation', 'cell_surface_expression'), n]

    def bp(self, n):
        """Return biological process."""
        return [self.nt_type('process', 'biological_process'), [n[0]]]

    def path(self, n):
        """Return path."""
        return [self.nt_type('process', 'pathology'), [n[0]]]

    def _format_sub_obj(self, obj):
        """Change namedtuples to tuple(name_of_namedtuple,dictionary)."""
        if isinstance(obj, (list, tuple)):
            for i in range(len(obj)):
                if not isinstance(obj[i], self.named_tuples) and isinstance(obj[i], (list, tuple)):
                    self._format_sub_obj(obj[i])
                elif isinstance(obj[i], self.named_tuples):
                    function_type = obj[i].__class__.__name__
                    if function_type == "function":
                        obj[i] = {function_type: obj[i]._asdict()}
                    else:
                        obj[i] = obj[i]._asdict()
        return obj

    def protein_changes(self, n):
        """Create dict of protein changes."""
        return self.changes(n)

    def statement_comment(self, n):
        return {'statement_comment': ' '.join([x.value for x in n]).strip()}

    def changes(self, n):
        """Create dict of abundance changes."""
        return [list((x[0], [x[1]])) for x in sorted(list(set(n)))]

    def gene_changes(self, n):
        """Return gene abundance changes; Only valid in BEL 2.1."""
        return [list((x[0], [x[1]])) for x in sorted(list(set(n)))]

    def gmod(self, n):
        """Return gmod; Only valid in BEL 2.1."""
        nn = [x for x in n if x.__class__.__name__ == 'nn']
        namespace = "TestNS"
        name = "TestName"

        if nn:
            namespace = nn[0].namespace
            name = nn[0].name
        nt = self.nt_gmod(
            namespace=namespace,
            name=name)

        return self.nt_type('modifier', 'gmod'), nt

    def ma(self, n):
        """Transform tokens to dict(function_name: namedtuple)."""
        namespace, name, default = (None,) * 3
        if type(n[0]).__name__ == 'nn':
            nt = n[0]
            namespace = nt.namespace
            name = nt.name

        elif n[0].data == 'ma_default':
            default = n[0].children[0].data

        return self.nt_ma(namespace=namespace, name=name, default=default)

    def frag(self, n) -> tuple:
        """Transform frag tokens to nametuple."""
        frag_range, frag_descriptor = self._get_values(n, ('F_RANGE', 'F_DESCRIPTOR'))
        nt = self.nt_frag(range=frag_range, descriptor=frag_descriptor)

        return self.nt_type('modifier', 'fragment'), nt

    def loc(self, n) -> tuple:
        """Transform tokens to dict(function_name: namedtuple)."""
        return self.nt_type('modifier', 'location'), n[0]

    def var(self, n) -> tuple:
        """Transform tokens to dict(function_name: namedtuple)."""
        return self.nt_type('modifier', 'variant'), self.nt_var(n[0].value)

    def nn(self, n):
        """Transform tokens to dict(function_name: namedtuple)."""
        token_names = ('NAMESPACE_KEYWORD', ('NAME_WITHOUT_QUOTES', 'STRING_SIMPLE'))
        namespace, name = self._get_values(n, token_names)

        token_dict = {x.type: x for x in n}
        name_with_quotes = token_dict.get('NAME_WITHOUT_QUOTES')
        simple_string = token_dict.get('STRING_SIMPLE')
        entry_token = name_with_quotes or simple_string

        self.cache.set_namespace_entry(
            namespace=namespace,
            entry=name,
            token=entry_token
        )

        return self.nt_nn(namespace=namespace, name=name)

    def pmod(self, n) -> dict:
        """Transform tokens to dict(function_name: namedtuple)."""
        namespace, name = '', ''

        pos_value = self._get_value(n, 'POSITION')
        position = int(pos_value) if pos_value else 0

        nn = [x for x in n if x.__class__.__name__ == 'nn']
        if nn:
            namespace = nn[0].namespace
            name = nn[0].name

        aa = self._get_dict_value(n, 'amino_acid')
        amino_acid = aa if aa else ''

        ptype = self._get_dict_value(n, 'pmod_type')
        type_ = ptype if ptype else ''

        nt = self.nt_pmod(
            namespace=namespace,
            name=name,
            type=type_,
            amino_acid=amino_acid,
            position=position)

        return self.nt_type('modifier', 'pmod'), nt

    def amino_acid(self, n):
        """Return amino acid."""
        return {'amino_acid': n[0].data.split('_')[1].upper()}

    def pmod_type(self, n):
        """Return pmod_type."""
        return {'pmod_type': n[0].data}

    def from_loc(self, n):
        """Return tloc."""
        return [self.nt_type('translocation', 'from_location'), n]

    def to_loc(self, n):
        """Return tloc."""
        return [self.nt_type('translocation', 'to_location'), n]

    def rxn(self, n):
        """Return reaction."""
        return [self.nt_type('transformation', 'reaction'), n]

    def reactants(self, n):
        """Return reactants."""
        return [self.nt_type('reaction_partner', 'reactants'), sorted(n)]

    def products(self, n):
        """Return products."""
        return [self.nt_type('reaction_partner', 'products'), sorted(n)]

    def fusion(self, n):
        """Return fusion."""
        return n[0]

    def _fusion_range(self, n):
        """Return fusion."""
        range_types = ('GENE_FUSION_RANGE', 'RNA_FUSION_RANGE', 'PROTEIN_FUSION_RANGE')
        fusion_range = self._get_value(n, range_types)
        return OrderedDict([('fusion_range', fusion_range)])

    def gene_fusion(self, n):
        """Return  gene fusion."""
        return [self.nt_type('other', 'fusion_gene'), n]

    def gene_fusion_range(self, n):
        """Return fusion range."""
        return self._fusion_range(n)

    def fus_gene1(self, n):
        return ['gene1', n]

    def fus_gene2(self, n):
        return ['gene2', n]

    def fus_rna1(self, n):
        return ['rna1', n]

    def fus_rna2(self, n):
        return ['rna2', n]

    def fus_protein1(self, n):
        return ['protein1', n]

    def fus_protein2(self, n):
        return ['protein2', n]

    def rna_fusion(self, n):
        """Return rna fusion."""
        return [self.nt_type('other', 'fusion_rna'), n]

    def rna_fusion_range(self, n):
        """Return fusion range."""
        return self._fusion_range(n)

    def protein_fusion(self, n):
        """Return protein fusion."""
        return [self.nt_type('other', 'fusion_protein'), n]

    def protein_fusion_range(self, n):
        """Return fusion range."""
        return self._fusion_range(n)

    def _get_dict_value(self, n, key_name):
        """Return value of key_name for first dict with key_name."""
        for e in n:
            if isinstance(e, dict) and key_name in e:
                return e[key_name]

    def __get_token_dict(self, tokens: typing.List[Token]) -> dict:
        """Get dictionary of tokens with typs as key."""
        return {t.type: t.value for t in tokens if isinstance(t, Token)}

    def _get_value(self, tokens, token_name):
        """Get first Token value of list with token_name.

        If token_name is a list method checks aginst list.
        """
        token_names = (token_name,) if isinstance(token_name, str) else token_name
        for t in tokens:
            if isinstance(t, Token) and t.type in token_names:
                return t.value

    def _get_values(self, tokens, token_names: typing.Iterable) -> list:
        """Get list of token values in same order then token_names.

        If token name not exists empty str is returned.
        If token_names element is a Iterable of str, first token value is
        return from this list.
        """
        t_dict = self.__get_token_dict(tokens)
        ret_lst = []

        for t_name in token_names:

            if isinstance(t_name, str):
                ret_lst.append(t_dict.get(t_name, ''))

            elif isinstance(t_name, typing.Iterable):
                value = ''
                for name in t_name:
                    value = t_dict.get(name)
                    if value:
                        break
                ret_lst.append(value)

        return ret_lst

    def _get_tree(self, n: list, tree_name):
        """Return first lark.Tree, namedtuple or dict linked to tree_name."""
        for e in n:
            if isinstance(e, Tree) and e.data == tree_name:
                return e
            elif isinstance(e, dict) and len(e.keys()) == 1:
                return e[tree_name]

        raise "Type not supported or empty list"

    def _get_all_values_by_name(self, tree_or_tokens, token_name):
        n = tree_or_tokens
        if isinstance(tree_or_tokens, Tree):
            n = tree_or_tokens.children

        return sorted([x.value for x in n if isinstance(x, Token) and x.type == token_name])

    def _first_tree(self, n: list) -> Tree:
        """Return first tree in list."""
        for e in n:
            if isinstance(e, Tree):
                return e



"""GWAS Catalog API methods."""
from flask import request

from ebel.manager.rdbms.models import gwas_catalog
from ebel.web.api.ebel.v1 import _get_data, _get_paginated_ebel_query_result


def get_gwas_catalog():
    """Get egenric GWAS catalog entry."""
    return _get_data(gwas_catalog.GwasCatalog)


def get_ebel_relation():
    """Get GWAS Catalog related eBEL relations."""
    allowed_relations = ['has_snp_gwascatalog',
                         'has_mapped_snp_gc',
                         'has_downstream_snp_gc',
                         'has_upstream_snp_gc']
    relation = request.args.get('relation')
    relation = relation if relation in allowed_relations else allowed_relations[0]

    conf = {
        'snp_rid': "in.@rid.asString()",
        'namespace': "out.namespace",
        'name': "out.name",
        'gene_rid': "out.@rid.asString()",
        'pubmed_id': "pubmed_id",
        'disease_trait': "disease_trait",
        'rs_number': "in.rs_number",
        'edge_rid': "@rid.asString()"
    }

    sql = "SELECT "
    sql += ', '.join([f"{v} as {k}" for k, v in conf.items()])
    sql += f" FROM {relation}"

    ra = request.args
    paras = {k: ra[k] for k in ra if k in conf}
    if paras:
        sql += " WHERE " + ' AND '.join([f'{conf[k]} like "{v}"' for k, v in paras.items()])

    return _get_paginated_ebel_query_result(sql, print_sql=True)



"""ClinVAR API unit tests."""

from .conftest import format_response_data
from .constants import RESULTS, NUM_RESULTS, PAGE_SIZE


class TestClinvar:

    def test_get_clinvar(self, client):
        response = client.get(
            'api/v1/clinvar?allele_id=705202&type=single%20nucleotide%20variant&name=NM_001772.4%28CD33%29%3Ac.131T%3EC%20%28p.Phe44Ser%29&gene_id=945&gene_symbol=CD33&hgnc_id=HGNC%3A1659&clinical_significance=Benign&last_evaluated=Jul%2016%2C%202018&rs_db_snp=61736469&rcvaccession=RCV000957991&origin=germline&origin_simple=germline&assembly=GRCh37&chromosome_accession=NC_000019.9&chromosome=19&start=51728567&stop=51728567&reference_allele=na&alternate_allele=na&cytogenetic=19q13.41&review_status=criteria%20provided%2C%20single%20submitter&number_submitters=1&tested_in_gtr=N&submitter_categories=2&variation_id=777507&position_vcf=51728567&reference_allele_vcf=T&alternate_allele_vcf=C&page_size=10&page=1',
            content_type='application/json'
        )

        output = format_response_data(response)

        expected_results = {
          "allele_id": 705202,
          "alternate_allele": "na",
          "alternate_allele_vcf": "C",
          "assembly": "GRCh37",
          "chromosome": "19",
          "chromosome_accession": "NC_000019.9",
          "clin_sig_simple": 0,
          "clinical_significance": "Benign",
          "cytogenetic": "19q13.41",
          "gene_id": 945,
          "gene_symbol": "CD33",
          "guidelines": None,
          "hgnc_id": "HGNC:1659",
          "id": 1132137,
          "last_evaluated": "Jul 16, 2018",
          "name": "NM_001772.4(CD33):c.131T>C (p.Phe44Ser)",
          "nsv_esv_db_var": None,
          "number_submitters": 1,
          "origin": "germline",
          "origin_simple": "germline",
          "otherIdentifiers": [],
          "phenotypeMedgens": [
            "CN517202"
          ],
          "position_vcf": 51728567,
          "rcvaccession": "RCV000957991",
          "reference_allele": "na",
          "reference_allele_vcf": "T",
          "review_status": "criteria provided, single submitter",
          "rs_db_snp": 61736469,
          "start": 51728567,
          "stop": 51728567,
          "submitter_categories": 2,
          "tested_in_gtr": "N",
          "type": "single nucleotide variant",
          "variation_id": 777507
        }

        assert output[NUM_RESULTS] == 1
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_clinvar_simple(self, client):
        response = client.get(
            'api/v1/clinvar/simple?phenotype=Alzheimer%20disease%202&hgnc_id=HGNC%3A4886&assembly=GRCh38&page_size=10&page=1',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_results = {
            "allele_id": 15048,
            "assembly": "GRCh38",
            "gene_symbol": "HFE",
            "hgnc_id": "HGNC:4886",
            "id": 14,
            "rs_db_snp": 1800562
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_phenotype_starts_with(self, client):
        response = client.get(
            'api/v1/clinvar/phenotype/starts_with?phenotype=MAN1B1-Related%20Disor&page_size=10&page=1',
            content_type='application/json'
        )
        output = format_response_data(response)
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, dict)
        assert "MAN1B1-Related Disorders" in results
        assert results["MAN1B1-Related Disorders"] >= 6899  # Value changes depending on when data was downloaded

    def test_get_by_other_identifier(self, client):
        response = client.get(
            'api/v1/clinvar/phenotype/by_other_identifier?db=OMIM&identifier=613653.%25&page_size=10&page=1',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_results = {
          "allele_id": 15041,
          "assembly": "GRCh37",
          "db": "OMIM",
          "gene_symbol": "AP5Z1",
          "hgnc_id": "HGNC:22197",
          "id": 1,
          "identifier": "613653.0001",
          "rs_db_snp": 397704705
        }
        assert output[NUM_RESULTS] > 10
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_by_medgen(self, client):
        response = client.get(
            'api/v1/clinvar/phenotype/by_medgen?identifier=C0002395&page_size=10&page=1',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_results = {
          "allele_id": 15048,
          "assembly": "GRCh37",
          "gene_symbol": "HFE",
          "hgnc_id": "HGNC:4886",
          "id": 13,
          "identifier": "C0002395",
          "rs_db_snp": 1800562
        }
        assert output[NUM_RESULTS] > 300
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results



"""Generic BEL API unit tests."""

from .constants import RESULTS
from .conftest import format_response_data


class TestBel:
    """Unit tests for the BEL API module."""

    def test_get_edge(self, client):
        response = client.get(
            'api/v1/bel/edges?subject_node_class=protein&subject_namespace=HGNC&subject_name=IL4&relation=increases&citation_full_journal_name=Molecular%20neurodegeneration&citation_pub_date=2018-06-12&citation_pub_year=2018&citation_last_author=Landreth%20GE&citation_type=PubMed&title=TREM2%20in%20Neurodegenerative%20Diseases.&doi=10.1186%2Fs13024-017-0197-5&object_node_class=protein&object_namespace=HGNC&object_name=TREM2&page_size=10&page=1',
            content_type='application/json'
        )
        cols = [
            'subject_rid',
            'subject_node_class',
            'subject_namespace',
            'subject_name',
            'subject_bel',
            'subject_gene_symbol_involved_in',
            'subject_other_involved_in',
            'edge_rid',
            'relation',
            'evidence',
            'citation_full_journal_name',
            'citation_pub_date',
            'citation_pub_year',
            'citation_last_author',
            'citation_type',
            'author_in_author_list',
            'title',
            'doi',
            'object_rid',
            'object_node_class',
            'object_namespace',
            'object_name',
            'object_bel',
            'object_gene_symbol_involved_in',
            'object_other_involved_in',
        ]

        output = format_response_data(response)
        assert isinstance(output, dict)
        results = output[RESULTS]
        assert isinstance(results, list)
        assert len(results) == 1
        hit = results[0]
        for col in cols:
            assert col in hit
        assert hit['object_name'] == "IL2"
        assert hit['object_namespace'] == "HGNC"
        assert hit['object_node_class'] == "protein"
        assert hit['citation_last_author'] == "Schellenberg GD"
        assert hit["relation"] == "increases"



"""CHEBI RDBMS model definition."""

from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, Text, ForeignKey, DateTime, Index

Base = declarative_base()


class ChemicalData(Base):
    """Class definition for the chebi_chemical_data table."""

    __tablename__ = 'chebi_chemical_data'
    id = Column(Integer, primary_key=True)

    chemical_data = Column(Text, nullable=True)
    source = Column(Text, nullable=False)
    type = Column(Text, nullable=False)

    compound_id = Column(Integer, ForeignKey('chebi_compound.id'))
    compounds = relationship("Compound", back_populates="chemicalData")

    def __str__(self):
        """Class string definition."""
        return self.chemical_data

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'chemical_data': self.chemical_data,
            'source': self.source,
            'type': self.type,
        }


class Comment(Base):
    """Class definition for the chebi_comment table."""

    __tablename__ = 'chebi_comment'
    id = Column(Integer, primary_key=True)

    text = Column(Text, nullable=False)
    created_on = Column(DateTime, nullable=False)
    datatype = Column(String(80))
    datatype_id = Column(Integer, nullable=False)

    compound_id = Column(Integer, ForeignKey('chebi_compound.id'))
    compounds = relationship("Compound", back_populates="comments")

    def __str__(self):
        """Class string definition."""
        return self.text

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'text': self.text,
            'datatype': self.datatype,
        }


class Compound(Base):
    """Class definition for the chebi_compound table."""

    __tablename__ = 'chebi_compound'
    id = Column(Integer, primary_key=True)

    name = Column(String(2000))
    source = Column(String(32), nullable=False)
    parent_id = Column(Integer)
    chebi_accession = Column(String(30), nullable=False)
    status = Column(String(1), nullable=False)
    definition = Column(Text)
    star = Column(Integer, nullable=False)
    modified_on = Column(Text)
    created_by = Column(Text)

    chemicalData = relationship("ChemicalData", back_populates="compounds")
    comments = relationship("Comment", back_populates="compounds")
    database_accessions = relationship("DatabaseAccession", back_populates="compounds")
    names = relationship("Name", back_populates="compounds")
    references = relationship("Reference", back_populates="compounds")
    # final_id_relations = relationship("Relation", back_populates="final_id_compounds")
    # init_id_relations = relationship("Relation", back_populates="init_id_compounds")
    structures = relationship("Structure", back_populates="compounds")
    inchis = relationship("Inchi", back_populates="compounds")

    def __str__(self):
        return self.name

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'name': self.name,
            'source': self.source,
            'parent_id': self.parent_id,
            'chebi_accession': self.chebi_accession,
            'status': self.status,
            'definition': self.definition,
            'chemicalData': [x.as_dict() for x in self.chemicalData],
            'comments': [x.as_dict() for x in self.comments],
            'database_accessions': [x.as_dict() for x in self.database_accessions],
            'names': [x.as_dict() for x in self.names],
            'references': [x.as_dict() for x in self.references],
            'inchis': [x.as_dict() for x in self.inchis]
        }


class Inchi(Base):
    """Class definition for the chebi_inchi table."""

    __tablename__ = 'chebi_inchi'
    id = Column(Integer, primary_key=True)

    inchi = Column(Text)

    compound_id = Column(Integer, ForeignKey('chebi_compound.id'))
    compounds = relationship("Compound", back_populates="inchis")

    def __str__(self):
        return self.inchi

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'inchi': self.inchi
        }


class DatabaseAccession(Base):
    """Class definition for the chebi_database_accession table."""

    __tablename__ = 'chebi_database_accession'
    id = Column(Integer, primary_key=True)

    accession_number = Column(String(255), nullable=True)
    type = Column(Text, nullable=False)
    source = Column(Text, nullable=False)

    compound_id = Column(Integer, ForeignKey('chebi_compound.id'))
    compounds = relationship("Compound", back_populates="database_accessions")

    def __str__(self):
        return self.accession_number

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'accession_number': self.accession_number,
            'type': self.type,
            'source': self.source,
        }


class Name(Base):
    """Class definition for the chebi_name table."""

    __tablename__ = 'chebi_name'
    id = Column(Integer, primary_key=True)

    name = Column(Text, nullable=True)
    type = Column(Text, nullable=False)
    source = Column(Text, nullable=False)
    adapted = Column(Text, nullable=False)
    language = Column(Text, nullable=False)

    compound_id = Column(Integer, ForeignKey('chebi_compound.id'))
    compounds = relationship("Compound", back_populates="names")

    def __str__(self):
        return self.name

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'name': self.name,
            'type': self.type,
            'source': self.source,
            'adapted': self.adapted,
            'language': self.language
        }


class Reference(Base):
    """Class definition for the chebi_reference table."""

    __tablename__ = 'chebi_reference'

    id = Column(Integer, primary_key=True)

    reference_id = Column(String(60), nullable=False, index=True)
    reference_db_name = Column(String(60), nullable=False, index=True)
    location_in_ref = Column(String(90), index=True)
    reference_name = Column(String(1024))

    compound_id = Column(Integer, ForeignKey('chebi_compound.id'))
    compounds = relationship("Compound", back_populates="references")

    __table_args__ = (Index('ix_chebi_reference__reference_name', reference_name, mysql_length=500),)

    def __str__(self):
        return f"{self.reference_db_name}:{self.reference_id}"

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'reference_id': self.reference_id,
            'reference_db_name': self.reference_db_name,
            'location_in_ref': self.location_in_ref,
            'reference_name': self.reference_name,
        }

    def as_dict_with_compound_id(self):
        """Convert object values to dictionary with compound ID."""
        return {
            'reference_id': self.reference_id,
            'reference_db_name': self.reference_db_name,
            'location_in_ref': self.location_in_ref,
            'reference_name': self.reference_name,
            'compound_id': self.compound_id
        }


class Relation(Base):
    """Class definition for the chebi_relation table."""

    __tablename__ = 'chebi_relation'
    id = Column(Integer, primary_key=True)

    type = Column(Text, nullable=False)
    status = Column(String(1), nullable=False)

    final_id = Column(Integer, ForeignKey('chebi_compound.id'))
    init_id = Column(Integer, ForeignKey('chebi_compound.id'))

    final_id_compounds = relationship("Compound", foreign_keys=[final_id])
    init_id_compounds = relationship("Compound", foreign_keys=[init_id])

    def __str__(self):
        return f"{self.type} - {self.status}"

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'type': self.type,
            'status': self.status,
            'final_id': self.final_id,
            'init_id': self.init_id
        }


class Structure(Base):
    """Class definition for the chebi_structure table."""

    __tablename__ = 'chebi_structure'
    id = Column(Integer, primary_key=True)

    structure = Column(Text, nullable=False)
    type = Column(Text, nullable=False)
    dimension = Column(Text, nullable=False)
    default_structure = Column(String(1), nullable=False)
    autogen_structure = Column(String(1), nullable=False)

    compound_id = Column(Integer, ForeignKey('chebi_compound.id'))
    compounds = relationship("Compound", back_populates="structures")

    def __str__(self):
        return self.structure



"""NCBI module."""
import pandas as pd

from tqdm import tqdm
from pyorientdb import OrientDB
from typing import Dict, Optional

from ebel.manager.orientdb.constants import NCBI
from ebel.manager.orientdb import odb_meta, urls
from ebel.tools import get_file_path, get_standard_name

from ebel.manager.rdbms.models import ncbi


class Ncbi(odb_meta.Graph):
    """NCBI Gene (https://ftp.ncbi.nih.gov/gene/DATA/)."""

    def __init__(self, client: OrientDB = None):
        """Init NcbiGene."""
        self.client = client
        self.biodb_name = NCBI
        self.urls = {ncbi.NcbiGeneInfo.__tablename__: urls.NCBI_GENE_INFO,
                     ncbi.NcbiGeneMim.__tablename__: urls.NCBI_GENE_2MIM,
                     ncbi.NcbiGeneEnsembl.__tablename__: urls.NCBI_GENE_2ENSEMBL,
                     ncbi.NcbiGeneGo.__tablename__: urls.NCBI_GENE_2GO,
                     ncbi.NcbiGenePubmed.__tablename__: urls.NCBI_GENE_2PUBMED,
                     ncbi.NcbiGeneOrtholog.__tablename__: urls.NCBI_GENE_ORTHOLOG,
                     ncbi.NcbiMedGenName.__tablename__: urls.NCBI_GENE_MEDGEN_NAMES,
                     ncbi.NcbiMedGenPmid.__tablename__: urls.NCBI_GENE_MEDGEN_PUBMED,
                     'neighbors': urls.NCBI_GENE_NEIGHBORS,
                     }
        super().__init__(urls=self.urls,
                         tables_base=ncbi.Base,
                         biodb_name=self.biodb_name)

    def __len__(self):
        return self.session.query(ncbi.NcbiGeneInfo).count()

    def __contains__(self, gene_id) -> bool:
        count = self.session.query(ncbi.NcbiGeneInfo).filter(ncbi.NcbiGeneInfo.gene_id == gene_id).count()
        return bool(count)

    def insert_data(self) -> Dict[str, int]:
        """Insert NCBI gene data.

        :return Dict[str, int]: Tables as keys and inserted rows and values.
        """
        self.recreate_tables()
        inserts = dict()
        inserts[ncbi.NcbiGeneInfo.__tablename__] = self._insert_info()
        inserts[ncbi.NcbiGeneGo.__tablename__] = self._insert_go()
        inserts[ncbi.NcbiGeneEnsembl.__tablename__] = self._insert_ensembl()
        # inserts[ncbi.NcbiGeneMim.__tablename__] = self._insert_mim()
        inserts[ncbi.NcbiGenePubmed.__tablename__] = self._insert_pubmeds()
        inserts[ncbi.NcbiGeneOrtholog.__tablename__] = self._insert_orthologs()
        inserts.update(self._insert_neighbors())
        inserts[ncbi.NcbiMedGenName.__tablename__] = self._insert_medgen()
        return inserts

    def __insert_table(self, model, use_cols=None, sep: str = "\t", chunksize: Optional[int] = None) -> int:
        """Generic method to insert data.

        :param model: NCBI SQLAlchemy model
        :param use_cols: Column names to be used in file
        :param sep: Column separator
        :return:
        """
        table = model.__tablename__
        file_path = get_file_path(self.urls[table], self.biodb_name)
        dfs = pd.read_csv(file_path, sep=sep, usecols=use_cols, chunksize=chunksize)
        if chunksize is None:
            dfs = [dfs]

        number_of_inserts = 0

        for df in dfs:
            self._standardize_dataframe(df)
            df.to_sql(table, self.engine, if_exists="append", index=False)
            number_of_inserts += df.shape[0]

        return number_of_inserts

    def _insert_medgen(self) -> int:
        """Insert MedGen names.

        :return: Number of inserts
        """
        table_name = ncbi.NcbiMedGenName.__tablename__
        file_path_name = get_file_path(self.urls[table_name], self.biodb_name)
        use_cols_name = ["#CUI", "name", "source", "SUPPRESS"]
        df_name = pd.read_csv(file_path_name, usecols=use_cols_name, sep='|').rename_axis('id')
        df_name.index += 1
        self._standardize_dataframe(df_name)
        df_name.to_sql(table_name, self.engine, if_exists="append")

        table_pmid = ncbi.NcbiMedGenPmid.__tablename__
        file_path_pmid = get_file_path(self.urls[table_pmid], self.biodb_name)
        use_cols_pmid = ["CUI", "PMID"]
        dfs_pmid = pd.read_csv(file_path_pmid, index_col='CUI', usecols=use_cols_pmid, sep='|', chunksize=1000000)

        df_name['ncbi_medgen_name_id'] = df_name.index
        df_name_fk = df_name.set_index('cui')[['ncbi_medgen_name_id']]

        inserted = 0
        for df in dfs_pmid:
            df.columns = [get_standard_name(x) for x in df.columns]
            df_join = df.join(df_name_fk).reset_index(drop=True)
            inserted += df_join.shape[0]
            df_join.to_sql(table_pmid, self.engine, if_exists="append", index=False)

        return inserted

    def _insert_go(self) -> int:
        """Insert Gene Ontology mapping to Gene IDs.

        :return: Number of inserts
        """
        table = ncbi.NcbiGeneGo.__tablename__
        file_path = get_file_path(self.urls[table], self.biodb_name)
        df = pd.read_csv(file_path, sep="\t")

        table_pmid = ncbi.NcbiGeneGoPmid.__tablename__

        self._standardize_dataframe(df)
        df.index += 1
        df.index.rename('id', inplace=True)
        df.pub_med = df.pub_med.str.split('|')
        rename_columns = {'id': "ncbi_gene_go_id", 'pub_med': "pmid"}
        df_pub_med = df[['pub_med']].explode('pub_med').dropna() \
            .reset_index().rename(columns=rename_columns)

        df.drop(columns=['pub_med'], inplace=True)
        df.to_sql(table, self.engine, if_exists="append")
        df_pub_med.to_sql(table_pmid, self.engine, if_exists="append", index=False)

        return df.shape[0]

    def _insert_ensembl(self) -> int:
        """Insert Ensembl mapping to Gene IDs.

        :return: Number of inserts
        """
        return self.__insert_table(ncbi.NcbiGeneEnsembl)

    def _insert_mim(self) -> int:
        """Insert OMIM (Online Mendelian Inheritance in Man) mapping to Gene IDs.

        :return: Number of inserts
        """
        return self.__insert_table(ncbi.NcbiGeneMim)

    def _insert_pubmeds(self) -> int:
        """Insert PubMed references to Gene IDs.

        :return: Number of inserts
        """
        return self.__insert_table(ncbi.NcbiGenePubmed, chunksize=1000000)

    def _insert_orthologs(self) -> int:
        """Insert orthologs to Gene IDs.

        :return: Number of inserts
        """
        use_cols = ['#tax_id', 'GeneID', 'Other_tax_id', 'Other_GeneID']
        return self.__insert_table(ncbi.NcbiGeneOrtholog, use_cols, chunksize=1000000)

    def _insert_neighbors(self) -> Dict[str, int]:
        """Insert overlapping, right and left neighbors.

        :return int: Number of inserts
        """
        usecols = ['GeneID', 'GeneIDs_on_left', 'GeneIDs_on_right', 'overlapping_GeneIDs']
        file_path = get_file_path(self.urls['neighbors'], self.biodb_name)

        neighbor_types = {'overlapping_gene_ids': ncbi.NcbiGeneOverlapping,
                          'gene_ids_on_right': ncbi.NcbiGeneOnRight,
                          'gene_ids_on_left': ncbi.NcbiGeneOnLeft}

        inserted = {neighbor_type: 0 for neighbor_type in neighbor_types}

        dfs = pd.read_csv(file_path, sep="\t", usecols=usecols, chunksize=1000000)

        for df in dfs:
            df.columns = [get_standard_name(x) for x in df.columns]

            for neighbor_type, model in neighbor_types.items():
                df_type = df[['gene_id', neighbor_type]].set_index('gene_id')
                ntype = neighbor_type.replace('ids', 'id')
                df_type.rename(columns={neighbor_type: ntype}, inplace=True)
                df_type[ntype] = df_type[ntype].str.split('|')
                df_type = df_type.explode(ntype)
                df_type = df_type[df_type[ntype] != '-'].reset_index()
                df_type.to_sql(
                    model.__tablename__,
                    self.engine,
                    if_exists='append',
                    index=False)
                inserted[neighbor_type] += df_type.shape[0]
        return inserted

    def _insert_info_description(self) -> pd.DataFrame:
        """Insert gene description.

        :return int: Number of inserts
        """
        table = ncbi.NcbiGeneInfo.__tablename__
        file_path = get_file_path(self.urls[table], self.biodb_name)
        df = pd.read_csv(file_path,
                         sep="\t",
                         usecols=['description'],
                         dtype={'description': 'string'}) \
            .dropna().drop_duplicates()
        df.index += 1
        df.index.rename('id', inplace=True)
        df.to_sql(ncbi.NcbiGeneInfoDescription.__tablename__, self.engine, if_exists='append')
        return df.assign(description_id=df.index).set_index(['description'])

    def _insert_info_xref(self, dataframe: pd.DataFrame) -> int:
        """Insert cross references to Gene IDs using Gene Info DataFrame.

        :param dataframe: Gene info pandas DataFrame
        :return int: Number of inserts
        """
        inserted = 0
        df = dataframe[['db_xrefs', 'gene_id']][dataframe.db_xrefs != '-'].dropna()

        if not df.empty:
            df.set_index('gene_id', inplace=True)
            df['db_xrefs'] = df['db_xrefs'].str.split('|')
            df = df.explode('db_xrefs')
            df[['db', 'dbid']] = df['db_xrefs'].str.split(':', 1, expand=True)
            df.drop(columns=['db_xrefs'], inplace=True)
            df.reset_index().to_sql('ncbi_gene_info_xref', self.engine, if_exists="append", index=False)
            inserted = df.shape[0]

        return inserted

    def _insert_info(self, chunksize: int = 1000000) -> int:
        """Insert Gene Info.

        :param chunksize: size of the row aggregates
        :return int: Number of inserts
        """
        df_info_decr = self._insert_info_description()
        inserted = 0
        table = ncbi.NcbiGeneInfo.__tablename__
        file_path = get_file_path(self.urls[table], self.biodb_name)

        # TODO: To decrease the size of gene_info table, type_of_gene should be implemented as its own table
        # df_type_of_gene = pd.read_csv(file_path, sep="\t",usecols=['type_of_gene']).drop_duplicates()\
        #     .reset_index(drop=True).rename_axis('id')
        # df_type_of_gene.index += 1
        # df_type_of_gene.to_sql()

        use_cols = {'#tax_id', 'GeneID', 'LocusTag', 'Symbol', 'chromosome', 'description', 'map_location',
                    'type_of_gene', 'dbXrefs'}
        for df in tqdm(pd.read_csv(file_path,
                                   sep="\t",
                                   low_memory=False,
                                   usecols=use_cols,
                                   chunksize=chunksize), desc=f"Import {self.biodb_name.upper()}"):
            self._standardize_dataframe(df)

            df.drop(columns=['db_xrefs']).set_index(['description']).join(df_info_decr).set_index('gene_id').to_sql(
                table, self.engine, if_exists="append", index=True)
            inserted += df.shape[0]

            self._insert_info_xref(df)

        return inserted

    def update_interactions(self) -> int:
        """Abstract method."""
        pass



"""Methods for interfacing with the RDBMS."""
import logging
import sys
import getpass
from typing import Optional, Union

import pymysql
from pyorientdb import OrientDB
from pyorientdb.exceptions import PyOrientCommandException, PyOrientConnectionException, \
    PyOrientSecurityAccessException

from ebel.defaults import CONN_STR_DEFAULT
from ebel.config import write_to_config
from ebel.constants import TerminalFormatting as TF

logger = logging.getLogger(__name__)


def orientdb_connection_works(server: str, port: int, name: str, user: str, password: str) -> bool:
    """Check if the connection to OrientDB works."""
    try:
        client = OrientDB(server, port)
        client.set_session_token(True)
        client.db_open(name, user, password)
        works = True

    except Exception as inst:
        print(type(inst))
        works = False

    return works


def get_orientdb_client(server: str, port: int, name: str, user: str, password: str,
                        root_password: Optional[str] = None, user_reader: Optional[str] = None,
                        user_reader_password: Optional[str] = None) -> OrientDB:
    """Attempts to connect to the OrientDB client. This is currently done by using session tokens."""
    # TODO PyOrientStorageException occurs with newer ODB versions
    client = OrientDB(server, port)

    # First try connect as admin user if this fails connect root_user from config,
    # if this fails ask for root password and create user with admin in database
    try:
        client.set_session_token(True)
        client.db_open(name, user, password)

    except (PyOrientCommandException, PyOrientConnectionException, PyOrientSecurityAccessException):
        client.set_session_token(True)
        root_passwd_correct = False
        while not root_passwd_correct:
            odb_root_question = f"{TF.QUESTION}OrientDB root password (to create database and users): {TF.RESET}"
            root_password = root_password or getpass.getpass(odb_root_question)
            try:
                client.connect('root', root_password)
                root_passwd_correct = True
            except (PyOrientConnectionException, PyOrientSecurityAccessException):
                logger.error(f'Connection problem to OrientDB server {server}:{port}')
                print(f"Please make sure the OrientDB server is running, port ({port}), "
                      f"as well server ({server}) and root password are correct")
                sys.exit()

        if not client.db_exists(name):
            client.db_create(name)
            logger.info(f"Create database '{name}'")
            client.db_open(name, 'root', root_password)
            # create user with admin rights
            client.command(
                f"CREATE USER {user} IDENTIFIED BY {password} ROLE admin")
            # create a reader
            if user_reader and user_reader_password:
                client.command(
                    f"CREATE USER {user_reader} IDENTIFIED BY {user_reader_password} ROLE reader")
            client.close()
            # reopen with new user and password
            client = OrientDB(server, int(port))
            client.set_session_token(True)
            client.db_open(name, user, password)
        else:
            client.db_open(name, 'root', root_password)
            # admin
            admin_user_exists_sql = "Select true as admin_user_exists from OUser " \
                                    f"where name = '{user}' and status='ACTIVE' and 'admin' in roles.name"
            admin_user_exists = client.command(admin_user_exists_sql)
            if admin_user_exists:
                print("Update password for OrientDB admin")
                client.command(
                    f"UPDATE OUser SET password = '{password}' WHERE name = '{user}'")
            else:
                print("Create password for OrientDB reader")
                client.command(
                    f"CREATE USER {user} IDENTIFIED BY {password} ROLE admin")

            # reader
            if user_reader and user_reader_password:
                reader_user_exists_sql = "Select true as admin_user_exists from OUser " \
                                         f"where name = '{user_reader}' and status='ACTIVE' and 'reader' in roles.name"
                reader_user_exists = client.command(reader_user_exists_sql)
                if reader_user_exists:
                    client.command(
                        f"UPDATE OUser SET password = '{user_reader_password}' WHERE name = '{user_reader}'")
                else:
                    client.command(
                        f"CREATE USER {user_reader} IDENTIFIED BY {user_reader_password} ROLE admin")

    return client


def set_mysql_interactive() -> tuple:
    """Interactive mode to setup MySQL database."""
    print("Interactive mode\n \
          ================\n \
          1st setup of db and user with root:\n")
    host = input("host[localhost]:") or 'localhost'
    user = input("ebel user[ebel]:") or 'ebel'
    password = getpass.getpass(prompt="ebel password[ebel]:") or 'ebel'
    db = input("database name[ebel]") or 'ebel'
    print("If you want to setup the database automatically,\n \
          then type in the root password, otherwise nothing")
    root_pwd = getpass.getpass(prompt='root password (only for 1st setup):')

    if root_pwd:
        root_host = getpass.getpass(prompt='IP or name mysql server [localhost]:') or 'localhost'
        conn = pymysql.connect(host=root_host, user='root', password=root_pwd)
        c = conn.cursor()
        db_exists = c.execute(f"show databases like '{db}'")

        if not db_exists:
            c.execute(f"CREATE DATABASE {db} CHARACTER SET utf8mb4 COLLATE utf8mb4_general_ci")
        else:
            logger.warning(f"Database '{db}' already exists!")

        user_exists = c.execute(f"Select 1 from mysql.user where User='{user}'")

        if not user_exists:
            sql = "CREATE USER '{}'@'{}' IDENTIFIED BY '{}'".format(
                user,
                host,
                password,
            )
            c.execute(sql)
        else:
            logger.warning(f"Database '{db}' already exists!")

        privileges_exists = c.execute(f"Select 1 from mysql.db where User='{user}' and Db='{db}'")
        if not privileges_exists:
            c.execute(f"GRANT ALL PRIVILEGES ON {db}.* TO '{user}'@'%'  IDENTIFIED BY '{password}'")
        else:
            logger.warning(f"User already has privileges for database '{db}'")

        c.execute("FLUSH PRIVILEGES")

    return host, user, password, db


def set_mysql_connection(host: str = 'localhost',
                         user: str = 'ebel_user',
                         password: str = 'ebel_passwd',
                         db: str = 'ebel',
                         port: Union[str, int] = '3306',
                         charset: str = 'utf8mb4'):
    """Set the connection using MySQL Parameters.

    Parameters
    ----------
    host : str
        MySQL database host.
    user : str
        MySQL database user.
    password : str
        MySQL database password.
    db : str
        MySQL database name.
    port : Union[str,int]
        The port of the MySQL database.
    charset : str
        MySQL database charset.

    Returns
    -------
    str
        SQLAlchemy MySQL connection string.

    """
    connection_string = f'mysql+pymysql://{user}:{password}@{host}/{db}:{port}?charset={charset}'
    set_connection(connection_string)

    return connection_string


def set_always_create_new_db(always_create_new_db: bool = True) -> None:
    """Set in configuration option `always_create_new_db`.

    Parameters
    ----------
    always_create_new_db : bool
        Option `always_create_new_db` in section `database` in config file.

    """
    write_to_config('database', 'always_create_new', str(always_create_new_db))


def set_connection(connection: str = CONN_STR_DEFAULT) -> None:
    """Set the connection string for SQLAlchemy.

    Parameters
    ----------
    connection: str
        SQLAlchemy connection string.

    """
    write_to_config('DATABASE', 'sqlalchemy_connection_string', connection)



"""OrientDB defaults."""
from enum import Enum
from typing import Dict


class OIndexType(Enum):
    """Allowed OrientDB Index types."""

    NOTUNIQUE = 'notunique'
    UNIQUE = 'unique'
    FULLTEXT = 'fulltext'
    DICTIONARY = 'dictionary'
    UNIQUE_HASH_INDEX = 'unique_hash_index'
    NOTUNIQUE_HASH_INDEX = 'notunique_hash_index'
    FULLTEXT_HASH_INDEX = 'fulltext_hash_index'
    DICTIONARY_HASH_INDEX = 'dictionary_hash_index'


class ODataType(Enum):
    """Allowed OrientDB Data types."""

    BOOLEAN = 'BOOLEAN'
    BINARY = 'BINARY'
    BYTE = 'BYTE'
    DATE = 'DATE'
    DATETIME = 'DATETIME'
    DECIMAL = 'DECIMAL'
    DOUBLE = 'DOUBLE'
    EMBEDDED = 'EMBEDDED'
    EMBEDDEDLIST = 'EMBEDDEDLIST'
    EMBEDDEDSET = 'EMBEDDEDSET'
    EMBEDDEDMAP = 'EMBEDDEDMAP'
    FLOAT = 'FLOAT'
    INTEGER = 'INTEGER'
    LONG = 'LONG'
    LINK = 'LINK'
    LINKLIST = 'LINKLIST'
    LINKSET = 'LINKSET'
    LINKMAP = 'LINKMAP'
    LINKBAG = 'LINKBAG'
    STRING = 'STRING'
    SHORT = 'SHORT'


bel_func_short: Dict[str, str] = {
    "gmod": 'gmod',
    "protein": 'p',
    "abundance": 'a',
    "micro_rna": 'm',
    "rna": 'r',
    "gene": 'g',
    "activity": 'act',
    "fragment": 'frag',
    "pmod": 'pmod',
    "location": 'loc',
    "variant": 'var',
    "complex": 'complex',
    "reaction": 'rxn',
    "reactants": 'reactants',
    "products": 'products',
    "pathology": 'path',
    "degradation": 'deg',
    "biological_process": 'bp',
    "list": 'list',
    "cell_secretion": 'sec',
    "composite": 'composite',
    "translocation": 'tloc',
    "fusion_protein": 'fus',
    "fusion_rna": 'fus',
    "fusion_gene": 'fus',
    "from_location": 'fromLoc',
    "to_location": 'toLoc',
    "cell_surface_expression": 'surf',
    "population": 'pop'
}

normalized_pmod: Dict[str, str] = {
    'ace': "Ac",
    'adr': "ADPRib",
    'add': "\"ADP-rybosylation\"",
    'far': "Farn",
    'ger': "Gerger",
    'gly': "Glyco",
    'hyd': "Hy",
    'isg': "ISG",
    'me0': "Me",
    'me1': "Me1",
    'mon': "monomethylation",
    'me2': "Me2",
    'me3': "Me3",
    'tri': "trimethylation",
    'myr': "Myr",
    'ned': "Nedd",
    'ngl': "NGlyco",
    'nit': "NO",
    'ogl': "OGlyco",
    'pal': "Palm",
    'pho': "Ph",
    'sul': "Sulf",
    'sup': "sulphation",
    'suh': "sulfonation",
    'sum': "sulphonation",
    'suy': "Sumo",
    'ubi': "Ub",
    'u48': "UbK48",
    'u63': "UbK63",
    'ubm': "UbMono",
    'ubp': "UbPoly",
    'pre': 'Prenylation',  # added for BioGrid
    'dei': 'de-ISGylation',  # added for BioGrid
    'fat': 'FAT10ylation'
}

normalized_pmod_reverse = {v: k for k, v in normalized_pmod.items()}


class BelPmod:
    """Protein modification definitions."""

    ACE = normalized_pmod['ace']
    ADR = normalized_pmod['adr']
    ADD = normalized_pmod['add']
    FAR = normalized_pmod['far']
    GER = normalized_pmod['ger']
    GLY = normalized_pmod['gly']
    HYD = normalized_pmod['hyd']
    ISG = normalized_pmod['isg']
    ME0 = normalized_pmod['me0']
    ME1 = normalized_pmod['me1']
    MON = normalized_pmod['mon']
    ME2 = normalized_pmod['me2']
    ME3 = normalized_pmod['me3']
    TRI = normalized_pmod['tri']
    MYR = normalized_pmod['myr']
    NED = normalized_pmod['ned']
    NGL = normalized_pmod['ngl']
    NIT = normalized_pmod['nit']
    OGL = normalized_pmod['ogl']
    PAL = normalized_pmod['pal']
    PHO = normalized_pmod['pho']
    SUL = normalized_pmod['sul']
    SUP = normalized_pmod['sup']
    SUH = normalized_pmod['suh']
    SUM = normalized_pmod['sum']
    SUY = normalized_pmod['suy']
    UBI = normalized_pmod['ubi']
    U48 = normalized_pmod['u48']
    U63 = normalized_pmod['u63']
    UBM = normalized_pmod['ubm']
    UBP = normalized_pmod['ubp']
    PRE = normalized_pmod['pre']  # added for BioGrid
    DEI = normalized_pmod['dei']  # added for BioGrid
    FAT = normalized_pmod['fat']



"""IUPHAR drug module."""

import logging
import numpy as np
import pandas as pd

from tqdm import tqdm
from typing import Dict
from pyorientdb import OrientDB

from ebel.tools import get_file_path
from ebel.manager.orientdb.constants import IUPHAR
from ebel.manager.orientdb.biodbs.uniprot import UniProt
from ebel.manager.orientdb import odb_meta, urls, odb_structure

from ebel.manager.rdbms.models import iuphar

logger = logging.getLogger(__name__)


class Iuphar(odb_meta.Graph):
    """IUPHAR."""

    def __init__(self, client: OrientDB = None):
        """Init IUPHAR."""
        self.client = client
        self.biodb_name = IUPHAR
        self.urls = {'iuphar_int': urls.IUPHAR_INT, 'iuphar_ligands': urls.IUPHAR_LIGANDS}
        super().__init__(edges=odb_structure.iuphar_edges,
                         tables_base=iuphar.Base,
                         urls=self.urls,
                         biodb_name=self.biodb_name)

    def __len__(self):
        return self.number_of_generics

    def __contains__(self, item):
        # TODO: To be implemented
        return True

    def insert_data(self) -> Dict[str, int]:
        """Insert data in generic OrientDB class."""
        ligands = self.insert_ligand_data()
        ints = self.insert_interaction_data()

        self.session.commit()
        return {'ligands': ligands, 'interactions': ints}

    def insert_ligand_data(self) -> int:
        """Insert ligand/drug data in generic OrientDB class."""
        logger.info("Reading and formatting IUPHAR drug data")
        df = pd.read_csv(get_file_path(self.urls['iuphar_ligands'], self.biodb_name),
                         sep=",",
                         low_memory=False,
                         skiprows=[0],
                         # dtype={'Ligand id': 'Int64', 'PubChem SID': 'Int64', 'PubChem CID': 'Int64'},
                         true_values=['yes']).replace({np.nan: None})  # Convert 'yes' to True
        df.columns = self._standardize_column_names(df.columns)
        df.rename(columns={'in_ch_i': 'inchi',
                           'in_ch_ikey': 'inchi_key',
                           'pub_chem_cid': 'pubchem_cid',
                           'pub_chem_sid': 'pubchem_sid',
                           'uni_prot_id': 'uniprot_id',
                           'ligand_id': 'id'}, inplace=True)  # Make it consistent with drugbank
        df.set_index('id', inplace=True)
        df.to_sql(iuphar.IupharLigand.__tablename__, self.engine, if_exists='append')

        return df.shape[0]

    def insert_interaction_data(self) -> int:
        """Insert interaction data in generic OrientDB class."""
        logger.info("Reading and formatting IUPHAR interaction data.")
        df = pd.read_csv(get_file_path(self.urls['iuphar_int'], self.biodb_name),
                         sep=",",
                         low_memory=False,
                         skiprows=[0],
                         dtype={'target_id': 'Int64', 'ligand_pubchem_sid': 'Int64', 'ligand_id': 'Int64',
                                'target_ligand_id': 'Int64', 'target_ligand_pub_chem_sid': 'Int64'},
                         true_values=['t'],
                         false_values=['f']).replace({np.nan: None})

        df.columns = self._standardize_column_names(df.columns)
        df.rename(columns={'target_uni_prot_id': "target_uniprot",
                           'target_ligand_uni_prot_id': "target_ligand_uniprot_id",
                           'target_ligand_pub_chem_sid': "target_ligand_pubchem_sid",
                           'ligand_pub_chem_sid': "ligand_pubchem_sid",
                           'pub_med_id': "pubmed_id"}, inplace=True)
        df.index += 1
        df.index.rename('id', inplace=True)
        df.to_sql(iuphar.IupharInteraction.__tablename__, self.engine, if_exists='append')

        return df.shape[0]

    def update_interactions(self) -> int:
        """Update the ligands/drug generic table to include a link to all interactions it is part of."""
        self.clear_edges()
        uniprot = UniProt(self.client)

        iuphar_edge_type_mapper = {
            'Agonist': 'agonist_of__iu',
            'Inhibitor': 'inhibits__iu',
            'Antagonist': 'antagonist_of__iu',
            'Channel blocker': 'channel_blocker_of__iu',
            'Allosteric modulator': 'allosteric_modulator_of__iu',
            'Activator': 'activates__iu',
            'Antibody': 'antibody_against__iu',
            'Gating inhibitor': 'inhibits_gating__iu'
        }

        sql = """select i.pubmed_id, i.assay_description, i.affinity_units, i.affinity_low, i.affinity_median,
        i.affinity_high, i.type,
        i.action,i.target_uniprot, l.name as ligand_name, l.pubchem_sid, i.ligand_gene_symbol, i.ligand_species
        from iuphar_interaction as i inner join iuphar_ligand as l
        on (i.ligand_id=l.id) where i.target_uniprot IS NOT NULL and pubchem_sid IS NOT NULL"""

        df_iuphar = pd.read_sql(sql, self.engine).replace({np.nan: None})
        df_iuphar.set_index('target_uniprot', inplace=True)
        df_graph = pd.DataFrame(
            uniprot.get_pure_uniprot_rid_dict_in_bel_context().items(),
            columns=['target_uniprot', 'rid'])
        df_graph.set_index('target_uniprot', inplace=True)
        df_join = df_graph.join(df_iuphar, how='inner')

        for uniprot, data in tqdm(df_join.iterrows(), total=df_join.shape[0],
                                  desc=f"Update {self.biodb_name.upper()} interactions"):
            if data.ligand_gene_symbol and data.ligand_species and 'Human' in data.ligand_species:
                symbol = data.ligand_gene_symbol.split('|')[0]  # human seems to always the first
                a_value_dict = {'pure': True,
                                'bel': f'p(HGNC:"{symbol}")',
                                'namespace': 'HGNC',
                                'name': symbol,
                                }
                a_class = 'protein'
            else:
                a_value_dict = {'pure': True,
                                'bel': f'a(PUBCHEM:"{data.pubchem_sid}")',
                                'namespace': 'PUBCHEM',
                                'name': str(data.pubchem_sid),
                                'label': data.ligand_name
                                }
                a_class = 'abundance'
            a_rid = self.get_create_rid(a_class, value_dict=a_value_dict, check_for='bel')

            i_value_dict = {'pmids': data.pubmed_id.split('|') if data.pubmed_id else None,
                            'assay_description': data.assay_description,
                            'affinity_units': data.affinity_units,
                            'affinity_low': data.affinity_low,
                            'affinity_median': data.affinity_median,
                            'affinity_high': data.affinity_high,
                            'type': data.type,
                            'action': data.action}
            edge_class = iuphar_edge_type_mapper.get(data.type, 'iuphar_interaction')
            self.create_edge(edge_class, from_rid=a_rid, to_rid=data.rid, value_dict=i_value_dict)

        # not sure if this is really needed
        # Hgnc(self.client).update_bel()

        return df_join.shape[0]
